from typing import Any, Literal

import jinja2
from pydantic import BaseModel


class Message(BaseModel):
    role: Literal["system", "user", "assistant"]
    content: str

    def render(self, **kwargs: str) -> dict[str, Any]:
        return {
            "role": self.role,
            "content": jinja2.Template(self.content).render(**kwargs),
        }

    class Config:
        # remove whitespace
        anystr_strip_whitespace = True


class ChatTemplate(BaseModel):
    messages: list[Message]

    def render(self, **kwargs: str) -> list[dict[str, Any]]:
        return [m.render(**kwargs) for m in self.messages]



from typing import Any, Callable

import pytest
from pydantic import BaseModel
from syrupy.assertion import SnapshotAssertion

from ..signature import FunctionSignature


def add(a: int, b: int = 10) -> int:  # type: ignore[empty-body]
    """Add two numbers."""


def concat(a: str, b: str) -> str:  # type: ignore[empty-body]
    """Concat two strings"""


class Person(BaseModel):
    """Person model."""

    name: str
    age: int


def fake_person(count: int) -> Person:  # type: ignore[empty-body]
    """generate fake person."""


def no_return_annoation(a: int, b: int):  # type: ignore[no-untyped-def]
    """Add two numbers."""


def concats(*args: str) -> str:  # type: ignore[empty-body]
    """Concat the given strings"""


def complex(a: str, b: str, *args: str, c: str, d: str, **kwargs: str) -> str:  # type: ignore[empty-body]
    """Complex function"""


@pytest.mark.parametrize(
    "func, args, kwargs",
    [
        (add, (1, 2), {}),
        (fake_person, (5,), {}),
        (concat, ("pen", "apple"), {}),
        (concat, ("a'b", "cd"), {}),
        (concats, ("a", "b", "c", "d"), {}),
        (
            complex,
            ("a-v", "b-v", "arg-v1", "arg-v2"),
            {"c": "c-v", "d": "d-v", "kwarg1": "kwarg1-v", "kwarg2": "kwarg2-v"},
        ),
    ],
)
def test_function_signature(
    func: Callable[[Any], Any],
    args: tuple[Any],
    kwargs: dict[str, Any],
    snapshot: SnapshotAssertion,
) -> None:
    sig = FunctionSignature(func)
    assert snapshot == sig.instruction()
    assert snapshot == sig.call_line(*args, **kwargs)


def test_function_signature_without_return_annoations() -> None:
    with pytest.raises(AssertionError):
        FunctionSignature(no_return_annoation)



from syrupy.assertion import SnapshotAssertion

from ..prompt import ChatTemplate, Message


def test_chat_template(snapshot: SnapshotAssertion) -> None:
    template = ChatTemplate(
        messages=[
            Message(role="system", content="hello, what's your name?"),
            Message(role="user", content="My name is {{name}}}"),
            Message(role="system", content="hello, {{name}} nice to meet you!"),
        ]
    )

    assert snapshot == template.render(name="John")



import json
import re
from typing import Generic, TypeVar

from pydantic import BaseModel, ValidationError
from pydantic.generics import GenericModel

T = TypeVar("T", bound=BaseModel)


class ParserError(Exception):
    pass


PYDANTIC_FORMAT_INSTRUCTIONS = """The output should be formatted as a JSON instance that conforms to the JSON schema below.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

Here is the output schema:
```
{schema}
```"""


class PydanticParser(GenericModel, Generic[T]):
    pydantic_model: type[T]

    def parse(self, text: str) -> T:
        try:
            # Greedy search for 1st json candidate.
            match = re.search(r"\{.*\}", text.strip(), re.MULTILINE | re.IGNORECASE | re.DOTALL)
            json_str = ""
            if match:
                json_str = match.group()
            json_object = json.loads(json_str)
            return self.pydantic_model.parse_obj(json_object)
        except (json.JSONDecodeError, ValidationError) as e:
            name = self.pydantic_model.__name__
            msg = f"Failed to parse {name} from completion {text}. Got: {e}"
            raise ParserError(msg)

    def get_format_instructions(self) -> str:
        schema = self.pydantic_model.schema()

        # Remove extraneous fields.
        reduced_schema = schema
        if "title" in reduced_schema:
            del reduced_schema["title"]
        if "type" in reduced_schema:
            del reduced_schema["type"]
        # Ensure json in context is well-formed with double quotes.
        schema_str = json.dumps(reduced_schema)

        return PYDANTIC_FORMAT_INSTRUCTIONS.format(schema=schema_str)



from functools import wraps
from typing import Any, Callable, ParamSpec, TypeVar

import openai

from .prompt import ChatTemplate, Message
from .utils.signature import FunctionSignature

T = TypeVar("T")
P = ParamSpec("P")


def ai_fn(
    fn: Callable[P, T],
) -> Callable[P, T]:
    sig = FunctionSignature(fn)

    @wraps(fn)
    def inner(*args: Any, **kwargs: Any) -> T:
        template = ChatTemplate(
            messages=[
                Message(role="system", content=sig.instruction()),
                Message(role="user", content=sig.call_line(*args, **kwargs)),
            ]
        )

        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=template.render(),
            temperature=0.0,
        )

        resp = response.choices[0].message["content"]
        return sig.parse(resp)

    return inner



from typing import Any

import pytest


def filter_response_headers(response: dict[str, Any]) -> dict[str, Any]:
    # Modify or exclude headers as needed
    response["headers"].pop("openai-organization", None)  # Exclude a header from the response
    return response


@pytest.fixture(scope="module")
def vcr_config() -> dict[str, Any]:
    return {
        "filter_headers": [
            "authorization",
        ],  # Be sure to match the case of the header exactly
        "before_record_response": filter_response_headers,
    }



import inspect
from typing import Any, Callable

import pydantic

from .pydantic_parser import PydanticParser


def clean_docstring(docstring: str) -> str:
    """Clean up docstring before sending to OpenAI."""
    output = []
    for line in docstring.split("\n"):
        output.append(line.strip())

    return "\n".join(output)


def format_call_line(func: Callable[..., Any], *args: Any, **kwargs: Any) -> str:
    args_list = [repr(arg) for arg in args]  # Convert arguments to their string representations
    kwargs_list = [f"{key}={repr(value)}" for key, value in kwargs.items()]  # Convert keyword arguments to string representations

    # Combine args and kwargs into a single list
    all_args = args_list + kwargs_list

    # Generate the formatted function call line
    call_line = f"{func.__name__}({', '.join(all_args)})"

    return call_line


class FunctionSignature:
    """A helper class to parse function signature and generate instruction for prompt."""

    def __init__(self, fn: Callable[..., Any]):
        self.fn = fn
        self.sig: inspect.Signature = inspect.signature(fn)
        return_annotation = self.sig.return_annotation

        assert return_annotation is not inspect.Signature.empty, f"Function {fn.__name__} must have return annotation"

        class UnpackModel(pydantic.BaseModel):
            ret: return_annotation  # type: ignore[valid-type]

        self.parser = PydanticParser[UnpackModel](pydantic_model=UnpackModel)

    def function_line(self) -> str:
        # NOTE: return type is instrcutive by parser
        f = str(self.sig.replace(return_annotation=inspect.Signature.empty))

        return f"def {self.fn.__name__}{f}:"

    def description(self) -> str:
        desc = clean_docstring(self.fn.__doc__ or "")

        return desc

    def call_line(self, *args: Any, **kwargs: Any) -> str:
        return format_call_line(self.fn, *args, **kwargs)

    def parse(self, text: str) -> Any:
        return self.parser.parse(text).ret

    def instruction(self) -> str:
        return clean_docstring(
            f"""You are now the following python function:
                    ```
                    # {self.description()}
                    {self.function_line()}
                    ```
                    Only respond with your `return` value.
                    {self.parser.get_format_instructions()}
                    """
        )



import pytest
from pydantic import BaseModel, Field
from syrupy.assertion import SnapshotAssertion

from ..pydantic_parser import ParserError, PydanticParser


class Receipt(BaseModel):
    """The Receipt for user"""

    amount: float
    currency: str = Field(description="ISO 4217 currency code")
    customer: str


def test_pydantic_parser_get_format_instructions(
    snapshot: SnapshotAssertion,
) -> None:
    assert snapshot == PydanticParser[Receipt](pydantic_model=Receipt).get_format_instructions()


def test_parse_output(snapshot: SnapshotAssertion) -> None:
    assert snapshot == PydanticParser[Receipt](pydantic_model=Receipt).parse('{"amount": 1.0, "currency": "USD", "customer": "John Doe"}')


def test_pydantic_parser_fail() -> None:
    with pytest.raises(ParserError):
        PydanticParser[Receipt](pydantic_model=Receipt).parse('{"amount": 1.0, "currency": "USD"}')
